/* blockgrep v0.0 
 * See LICENSE file for details.
 * Ivan S. Jrgensen
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <regex.h>
#include <errno.h>
#include <unistd.h>

static regex_t *pattern_reg=0;
static int patterns=0;
static regex_t separator_reg;

/* See if the block matches. If it does then print it */
static void processBlock(const char *block, size_t block_bytes) {
	int match=0;
	int patnum;
	for(patnum=0; patnum<patterns && !match; patnum++) {
		if(regexec(&pattern_reg[patnum], block, 0,0, 0)==0)
			match=1;
	}
	if(match)
		fwrite(block,1,block_bytes,stdout);
}


/*Process standard input separating into blocks*/
static void processStdin(size_t max_block_size) {
	char *current_block = 0;
	size_t current_block_bytes=0;
	size_t current_block_size=0;
	char line[4096];
	
	setvbuf(stdin, 0, _IOFBF, 32678);
	
	while(fgets(line,sizeof(line),stdin)) {
		size_t linelen = strlen(line);
		if(linelen!=0 && line[linelen-1]=='\n') line[linelen--]='\0';
		if(regexec(&separator_reg, line, 0,0, 0)==0) {
			if(current_block) {
				current_block[current_block_bytes]='\0';
				processBlock(current_block,current_block_bytes);
				current_block_bytes=0;
			}
		}
		if(current_block_bytes+linelen+2 > current_block_size) {
			size_t new_block_size = current_block_bytes+linelen+2;
			if(new_block_size>max_block_size) {
				fprintf(stderr,"blockgrep: Maximum block size reached\n");
				exit(EXIT_FAILURE);
			}
			new_block_size = (new_block_size+4095)&~4096;
			current_block = (char*)realloc(current_block,new_block_size);
			if(!current_block) {
				fprintf(stderr,"blockgrep: failed to expand block buffer to %ld bytes, errno=%d\n", (long)new_block_size, errno);
				exit(EXIT_FAILURE);
			}
			current_block_size = new_block_size;
		}
		line[linelen++]='\n';
		memcpy(current_block+current_block_bytes, line, linelen);
		current_block_bytes += linelen;
	}
	if(current_block_bytes) {
		current_block[current_block_bytes]='\0';
		processBlock(current_block,current_block_bytes);
	}
	free(current_block);
}


static void printUsage(void) {
	fprintf(stderr,"Name: blockgrep\n");
	fprintf(stderr,"\n");
	fprintf(stderr,"Synopsis:\n");
	fprintf(stderr,"  blockgrep [OPTION] SEPARATOR [PATTERN...]\n");
	fprintf(stderr,"\n");
	fprintf(stderr,"Description:\n");
	fprintf(stderr,"  Splits standard input into blocks based on SEPARATOR and outputs\n");
	fprintf(stderr,"  blocks that matches any of the specified PATTERNS\n");
	fprintf(stderr,"  Both SEPARATOR and PATTERN are basic regular expressions (BRE) or\n");
	fprintf(stderr,"  extended regular expressions (ERE)\n");
	fprintf(stderr,"\n");
	fprintf(stderr,"Options:\n");
	fprintf(stderr,"  -b bytes  Set maximum block size. Blocks larger than this will make\n");
	fprintf(stderr,"            blockgrep exit. Default size is 128KB.\n");
	fprintf(stderr,"  -E        Interpret patterns/separator as ERE instead of BRE\n");
	fprintf(stderr,"  -G        Interpret patterns/separator as BRE instead of ERE (default)\n");
	fprintf(stderr,"  -i        Pattern is case-insensitive. Default is case-sensitive.\n");
	fprintf(stderr,"  -I        Separator is case-insensitive. Default is case-sensitive.\n");
	fprintf(stderr,"  -V        Print version number and exit\n");
	fprintf(stderr,"\n");
	fprintf(stderr,"Bugs:\n");
	fprintf(stderr,"  Does not handle start-of-line and end-of-line in regular expressions\n");
}


int main(int argc, char **argv) {
	int optchar;
	int rcrc;
	int cflags_separator=REG_NOSUB;
	int cflags_pattern=REG_NOSUB;
	size_t max_block_size = 128*1024;
	int patidx;
	int argidx;
	while((optchar=getopt(argc,argv,"b:EGiIV"))!=-1) {
		switch(optchar) {
			case 'b': {
				char *endptr=0;
				unsigned long l = strtoul(optarg,&endptr,0);
				if(endptr && *endptr!='\0') {
					fprintf(stderr,"blockgrep: Could not parse -b argument '%s'\n",optarg);
					return EXIT_FAILURE;
				}
				max_block_size = (size_t)l;
				break;
			}
			case 'E':
				cflags_separator |= REG_EXTENDED;
				cflags_pattern   |= REG_EXTENDED;
				break;
			case 'G':
				cflags_separator &= ~(REG_EXTENDED);
				cflags_pattern   &= ~(REG_EXTENDED);
				break;
			case 'i':
				cflags_pattern |= REG_ICASE;
				break;
			case 'I':
				cflags_separator |= REG_ICASE;
				break;
			case 'V':
				fprintf(stderr,"blockgrep 0.0\n");
				return EXIT_FAILURE;
			case '?':
			default:
				fprintf(stderr,"blockgrep: usage error\n");
				printUsage();
				return EXIT_FAILURE;
		}
	}
	
	if(optind==argc) {
		fprintf(stderr,"blockgrep: usage error: no separator\n");
		return EXIT_FAILURE;
	}
	
	rcrc = regcomp(&separator_reg,argv[optind],cflags_separator);
	if(rcrc!=0) {
		char errbuf[128];
		regerror(rcrc, &separator_reg, errbuf, sizeof(errbuf));
		fprintf(stderr,"blockgrep: separator: %s\n",errbuf);
		return EXIT_FAILURE;
	}
	
	patterns = argc-optind-1;
	pattern_reg = (regex_t*)calloc(patterns,sizeof(regex_t));
	patidx=0;
	for(argidx=optind+1; argidx<argc; argidx++) {
		rcrc = regcomp(&pattern_reg[patidx], argv[argidx], cflags_pattern);
		if(rcrc!=0) {
			char errbuf[128];
			regerror(rcrc, &pattern_reg[patidx], errbuf, sizeof(errbuf));
			fprintf(stderr,"blockgrep: pattern: %s\n",errbuf);
			return EXIT_FAILURE;
		}
		patidx++;
	}
	
	processStdin(max_block_size);
	
	return EXIT_SUCCESS;
}
